@@ -48,8 +48,8 @@ char copyright[] =
  * or by the name of the daemon. If command-line arguments are present, they 
  * take priority. The options are:
  * -k means trust krb4 or krb5
-* -5 means trust krb5
-* -4 means trust krb4 (using .klogin)
+ * -5 means trust krb5
+ * -4 means trust krb4 (using .klogin)
  * 
  */
      
@@ -73,7 +73,7 @@ char copyright[] =
 #define SERVE_NON_KRB     
 #define LOG_REMOTE_REALM
 #define LOG_CMD
-     
+   
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -87,7 +87,10 @@ char copyright[] =
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/param.h>
+#if !defined(KERBEROS) || !defined(KRB5_KRB4_COMPAT)
+/* Ultrix doesn't protect it vs multiple inclusion, and krb.h includes it */
 #include <sys/socket.h>
+#endif
 #include <sys/file.h>
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -115,8 +118,11 @@ char copyright[] =
 #endif
      
 #include <signal.h>
+#if !defined(KERBEROS) || !defined(KRB5_KRB4_COMPAT)
+/* Ultrix doesn't protect it vs multiple inclusion, and krb.h includes it */
 #include <netdb.h>
-     
+#endif
+
 #ifdef CRAY
 #ifndef NO_UDB
 #include <udb.h>
@@ -146,42 +152,49 @@ char copyright[] =
 #endif
 
 #ifdef KERBEROS
-#include "krb5.h"
-#include "com_err.h"
+#include <krb5.h>
+#include <com_err.h>
 #include "loginpaths.h"
+#ifdef KRB5_KRB4_COMPAT
+#include <kerberosIV/krb.h>
+Key_schedule v4_schedule;
+#endif
+
+#include "defines.h"
+
+#if HAVE_ARPA_NAMESER_H
+#include <arpa/nameser.h>
+#endif
+
+#ifndef MAXDNAME
+#define MAXDNAME 256 /*per the rfc*/
+#endif
+
+#define ARGSTR	"ek54ciD:S:M:AP:?L:w:"
 
-#define ARGSTR	"ek54ciD:S:M:AP:?L:"
 
 
-#define RSHD_BUFSIZ 5120
 
 #define MAXRETRIES 4
 
-char des_inbuf[2*RSHD_BUFSIZ];    /* needs to be > largest read size */
-krb5_encrypt_block eblock;        /* eblock for encrypt/decrypt */
-char des_outbuf[2*RSHD_BUFSIZ];   /* needs to be > largest write size */
-krb5_data desinbuf,desoutbuf;
 krb5_context bsd_context;
 char *srvtab = NULL;
 krb5_keytab keytab = NULL;
 krb5_ccache ccache = NULL;
 void fatal();
-int v5_des_read();
-int v5_des_write();
 
-int (*des_read)() = v5_des_read;
-int (*des_write)() = v5_des_write;
 int require_encrypt = 0;
 int do_encrypt = 0;
 int anyport = 0;
 char *kprogdir = KPROGDIR;
 int netf;
+int maxhostlen = 0;
+int stripdomain = 1;
+int always_ip = 0;
 
 #else /* !KERBEROS */
 
 #define ARGSTR	"RD:?"
-int (*des_read)() = read;
-int (*des_write)() = write;
      
 #endif /* KERBEROS */
      
@@ -281,7 +294,7 @@ int main(argc, argv)
     
     /* Analyze parameters. */
     opterr = 0;
-    while ((ch = getopt(argc, argv, ARGSTR)) != EOF)
+    while ((ch = getopt(argc, argv, ARGSTR)) != -1)
       switch (ch) {
 #ifdef KERBEROS
 	case 'k':
@@ -349,8 +362,30 @@ int main(argc, argv)
 	case 'D':
 	  debug_port = atoi(optarg);
 	  break;
-	case '?':
-	default:
+      case 'w':
+	  if (!strcmp(optarg, "ip"))
+	      always_ip = 1;
+	  else {
+	      char *cp;
+	      cp = strchr(optarg, ',');
+	      if (cp == NULL)
+		  maxhostlen = atoi(optarg);
+	      else if (*(++cp)) {
+		  if (!strcmp(cp, "striplocal"))
+		      stripdomain = 1;
+		  else if (!strcmp(cp, "nostriplocal"))
+		      stripdomain = 0;
+		  else {
+		      usage();
+		      exit(1);
+		  }
+		  *(--cp) = '\0';
+		  maxhostlen = atoi(optarg);
+	      }
+	  }
+	  break;
+      case '?':
+      default:
 	  usage();
 	  exit(1);
 	  break;
@@ -443,8 +478,10 @@ char    term[64] = "TERM=network";
 char	path_rest[] = RPATH;
 
 char	remote_addr[64];	/* = "KRB5REMOTEADDR=" */
+char	remote_port[64];	/* = "KRB5REMOTEPORT=" */
 char	local_addr[64];		/* = "KRB5LOCALADDR=" */
-#define ADDRPAD 0,0		/* remoteaddr, localaddr */
+char	local_port[64];		/* = "KRB5LOCALPORT=" */
+#define ADDRPAD 0,0,0,0
 #define KRBPAD 0		/* KRB5CCNAME, optional */
 
 /* The following include extra space for TZ and MAXENV pointers... */
@@ -492,9 +529,10 @@ char *kremuser;
 krb5_principal client;
 krb5_authenticator *kdata;
 
-#include <kerberosIV/krb.h>
+#ifdef KRB5_KRB4_COMPAT
 AUTH_DAT	*v4_kdata;
 KTEXT		v4_ticket;
+#endif
 
 int auth_sys = 0;	/* Which version of Kerberos used to authenticate */
 
@@ -572,11 +610,13 @@ void doit(f, fromp)
     
     int s;
     struct hostent *hp;
-    char *hostname;
+    char hostname[MAXDNAME];
+    char *sane_host;
+    char hostaddra[16];
     short port;
     int pv[2], pw[2], px[2], cc;
     fd_set ready, readfrom;
-    char buf[RSHD_BUFSIZ], sig;
+    char buf[RCMD_BUFSIZ], sig;
     struct sockaddr_in fromaddr;
     struct sockaddr_in localaddr;
     int non_privileged = 0;
@@ -601,7 +641,7 @@ void doit(f, fromp)
     
     { 
       int sin_len = sizeof (struct sockaddr_in);
-      if (getsockname(f, &localaddr, &sin_len) < 0) {
+      if (getsockname(f, (struct sockaddr*)&localaddr, &sin_len) < 0) {
 	perror("getsockname");
 	exit(1);
       }
@@ -635,8 +675,6 @@ void doit(f, fromp)
     }
 #ifdef KERBEROS
     netf = f;
-    desinbuf.data = des_inbuf;
-    desoutbuf.data = des_outbuf;
     if ( (fromp->sin_port >= IPPORT_RESERVED ||
 	    fromp->sin_port < IPPORT_RESERVED/2))
       non_privileged = 1;
@@ -698,12 +736,10 @@ void doit(f, fromp)
     }
     (void) alarm(0);
     if (port != 0) {
-	int lport;
 	if (anyport) {
-	    lport = 5120; /* arbitrary */
-	    s = getport(&lport);
+	    s = getport(0);
 	} else {
-	    lport = IPPORT_RESERVED - 1;
+	    int lport = IPPORT_RESERVED - 1;
 	    s = rresvport(&lport);
 	}
 	if (s < 0) {
@@ -713,7 +749,7 @@ void doit(f, fromp)
 	}
 #ifdef KERBEROS
 	if ( port >= IPPORT_RESERVED)
-	  non_privileged = 1;
+	    non_privileged = 1;
 #else
 	if (port >= IPPORT_RESERVED) {
 	    syslog(LOG_ERR , "2nd port not reserved\n");
@@ -732,26 +768,32 @@ void doit(f, fromp)
     dup2(f, 2);
     hp = gethostbyaddr((char *)&fromp->sin_addr, sizeof (struct in_addr),
 		       fromp->sin_family);
-    if (hp){
-	hostname = malloc(strlen(hp->h_name) + 1);
-	strcpy(hostname,hp->h_name);
-    }
-    else {
-	hostname = malloc(strlen((char *)inet_ntoa(fromp->sin_addr)) + 1);
-	strcpy(hostname,(char *)inet_ntoa(fromp->sin_addr));
+    strncpy(hostaddra, inet_ntoa(fromp->sin_addr), sizeof (hostaddra));
+    hostaddra[sizeof (hostaddra) - 1] = '\0';
+    if (hp != NULL){
+	strncpy(hostname, hp->h_name, sizeof (hostname));
+	hostname[sizeof (hostname) - 1] = '\0';
     }
+    else
+	hostname[0] = '\0';
 
 #ifdef KERBEROS
+    status = pty_make_sane_hostname(fromp, maxhostlen,
+				    stripdomain, always_ip, &sane_host);
+    if (status) {
+	error("failed make_sane_hostname: %s\n", error_message(status));
+	exit(1);
+    }
+
     if ((status = recvauth(f, fromaddr,&valid_checksum))) {
 	error("Authentication failed: %s\n", error_message(status));
 	exit(1);
     }
-    if (!strncmp(cmdbuf, "-x ", 3))
-	do_encrypt = 1;
 #else
     getstr(f, remuser, sizeof(remuser), "remuser");
     getstr(f, locuser, sizeof(locuser), "locuser");
     getstr(f, cmdbuf, sizeof(cmdbuf), "command");
+    rcmd_stream_init_normal();
 #endif /* KERBEROS */
     
 #ifdef CRAY
@@ -773,8 +815,9 @@ void doit(f, fromp)
     pwd = getpwnam(locuser);
     if (pwd == (struct passwd *) 0 ) {
 	syslog(LOG_ERR ,
-	       "Principal %s (%s@%s) for local user %s has no account.\n",
-	       kremuser, remuser, hostname, locuser); /* xxx sprintf buffer in syslog*/
+	       "Principal %s (%s@%s (%s)) for local user %s has no account.\n",
+	       kremuser, remuser, hostaddra, hostname,
+	       locuser); /* xxx sprintf buffer in syslog*/
 	error("Login incorrect.\n");
 	exit(1);
     }
@@ -825,18 +868,18 @@ void doit(f, fromp)
     endudb();
     if (secflag) {
 	if(getsysv(&sysv, sizeof(struct sysv)) != 0) {
-	    loglogin(hostname, SLG_LLERR, 0, ue);
+	    loglogin(sane_host, SLG_LLERR, 0, ue);
 	    error("Permission denied.\n");
 	    exit(1);
 	}
 	if ((packet_level != ue->ue_deflvl) ||
 	    ((packet_compart & ue->ue_comparts) != packet_compart )){
-	    loglogin(hostname, SLG_LLERR, 0, ue);
+	    loglogin(sane_host, SLG_LLERR, 0, ue);
 	    error("Permission denied.\n");
 	    exit(1);
 	}
 	if (ue->ue_disabled != 0) {
-	    loglogin(hostname,SLG_LOCK,ue->ue_logfails,ue);
+	    loglogin(sane_host,SLG_LOCK,ue->ue_logfails,ue);
 	    error("Permission denied.\n");
 	    exit(1);
 	}
@@ -870,13 +913,13 @@ void doit(f, fromp)
     if (port) {
 	/* Place entry into wtmp */
 	sprintf(ttyn,"krsh%1d",getpid());
-	pty_logwtmp(ttyn,locuser,hostname);
+	pty_logwtmp(ttyn,locuser,sane_host);
     }
     /*      We are simply execing a program over rshd : log entry into wtmp,
 	    as kexe(pid), then finish out the session right after that.
 	    Syslog should have the information as to what was exec'd */
     else {
-	pty_logwtmp(ttyn,locuser,hostname);
+	pty_logwtmp(ttyn,locuser,sane_host);
     }
     
 #ifdef CRAY
@@ -889,7 +932,7 @@ void doit(f, fromp)
 	if (getusrv(&usrv)){
 	    syslog(LOG_ERR,"Cannot getusrv");
 	    error("Permission denied.\n");
-	    loglogin(hostname, SLG_LVERR, ue->ue_logfails,ue);
+	    loglogin(sane_host, SLG_LVERR, ue->ue_logfails,ue);
 	    goto signout_please;
 	}
 	/*
@@ -898,12 +941,12 @@ void doit(f, fromp)
 	if((ue->ue_valcat & TFM_TRUSTED) ||
 	   (sysv.sy_oldtfm &&
 	    ((ue->ue_comparts & TRUSTED_SUBJECT) == TRUSTED_SUBJECT))) {
-	    loglogin(hostname, SLG_TRSUB, ue->ue_logfails,ue);
+	    loglogin(sane_host, SLG_TRSUB, ue->ue_logfails,ue);
 	    error("Permission denied.\n");
 	    goto signout_please;
 	}
 	
-	loglogin(hostname, SLG_OKLOG, ue->ue_logfails,ue);
+	loglogin(sane_host, SLG_OKLOG, ue->ue_logfails,ue);
 	
 	/*	Setup usrv structure with user udb info and 
 		packet_level and packet_compart. */
@@ -988,7 +1031,7 @@ void doit(f, fromp)
 	    }
 	}
 	if (nal_error) {
-	    loglogin(hostname, SLG_LVERR, ue->ue_logfails,ue);
+	    loglogin(sane_host, SLG_LVERR, ue->ue_logfails,ue);
 	    error("Permission denied.\n");
 	    goto signout_please;
 	}
@@ -998,7 +1041,7 @@ void doit(f, fromp)
 	sethost(paddr);
 	
 	if (setusrv(&usrv) == -1) {
-	    loglogin(hostname, SLG_LVERR, ue->ue_logfails,ue);
+	    loglogin(sane_host, SLG_LVERR, ue->ue_logfails,ue);
 	    error("Permission denied.\n");
 	    goto signout_please;
 	}
@@ -1025,8 +1068,8 @@ void doit(f, fromp)
 	    /* kuserok returns 0 if OK */
 	    if (kuserok(v4_kdata, locuser)){
 		syslog(LOG_ERR ,
-		       "Principal %s (%s@%s) for local user %s failed kuserok.\n",
-		       kremuser, remuser, hostname, locuser);
+		       "Principal %s (%s@%s (%s)) for local user %s failed kuserok.\n",
+		       kremuser, remuser, hostaddra, hostname, locuser);
 		}
 	    else auth_sent |= AUTH_KRB4;
 	} else
@@ -1035,8 +1078,8 @@ void doit(f, fromp)
 	    /* krb5_kuserok returns 1 if OK */
 	    if (!krb5_kuserok(bsd_context, client, locuser)){
 		syslog(LOG_ERR ,
-		       "Principal %s (%s@%s) for local user %s failed krb5_kuserok.\n",
-		       kremuser, remuser, hostname, locuser);
+		       "Principal %s (%s@%s (%s)) for local user %s failed krb5_kuserok.\n",
+		       kremuser, remuser, hostaddra, hostname, locuser);
 	    }
 	    else
 		auth_sent |=
@@ -1046,7 +1089,8 @@ void doit(f, fromp)
 	
 #else
     if (pwd->pw_passwd != 0 && *pwd->pw_passwd != '\0' &&
-	ruserok(hostname, pwd->pw_uid == 0, remuser, locuser) < 0) {
+	ruserok(hostname[0] ? hostname : hostaddra,
+		pwd->pw_uid == 0, remuser, locuser) < 0) {
 	error("Permission denied.\n");
 	goto signout_please;
     }
@@ -1084,11 +1128,11 @@ void doit(f, fromp)
     pwd = (struct passwd *) getpwnam(locuser);
     if (pwd && (pwd->pw_uid == 0)) {
 #ifdef LOG_CMD
-	syslog(LOG_NOTICE, "Executing %s for principal %s (%s@%s) as ROOT", 
-	       cmdbuf, kremuser, remuser, hostname);
+	syslog(LOG_NOTICE, "Executing %s for principal %s (%s@%s (%s)) as ROOT", 
+	       cmdbuf, kremuser, remuser, hostaddra, hostname);
 #else
-	syslog(LOG_NOTICE ,"Access as ROOT by principal %s (%s@%s)",
-	       kremuser, remuser, hostname);
+	syslog(LOG_NOTICE ,"Access as ROOT by principal %s (%s@%s (%s))",
+	       kremuser, remuser, hostaddra, hostname);
 #endif
     }
 #if defined(KERBEROS) && defined(LOG_REMOTE_REALM) && !defined(LOG_OTHER_USERS) && !defined(LOG_ALL_LOGINS)
@@ -1108,11 +1152,11 @@ void doit(f, fromp)
 #if defined(LOG_REMOTE_REALM) || defined(LOG_OTHER_USERS) || defined(LOG_ALL_LOGINS)
       {
 #ifdef LOG_CMD
-	  syslog(LOG_NOTICE, "Executing %s for principal %s (%s@%s) as local user %s", 
-		 cmdbuf, kremuser, remuser, hostname, locuser);
+	  syslog(LOG_NOTICE, "Executing %s for principal %s (%s@%s (%s)) as local user %s", 
+		 cmdbuf, kremuser, remuser, hostaddra, hostname, locuser);
 #else
-	  syslog(LOG_NOTICE ,"Access as %s by principal %s (%s@%s)",
-		 locuser, kremuser, remuser, hostname);
+	  syslog(LOG_NOTICE ,"Access as %s by principal %s (%s@%s (%s))",
+		 locuser, kremuser, remuser, hostaddra, hostname);
 #endif
       }
 #endif
@@ -1201,7 +1245,7 @@ if(port)
 			shutdown(s, 1+1);
 			FD_CLR(pv[0], &readfrom);
 		    } else {
-			(void) (*des_write)(s, buf, cc);
+			(void) rcmd_stream_write(s, buf, cc, 1);
 		    }
 		}
 		if (FD_ISSET(pw[0], &ready)) {
@@ -1212,12 +1256,12 @@ if(port)
 			shutdown(f, 1+1);
 			FD_CLR(pw[0], &readfrom);
 		    } else {
-			(void) (*des_write)(f, buf, cc);
+			(void) rcmd_stream_write(f, buf, cc, 0);
 		    }
 		}
 		if (port&&FD_ISSET(s, &ready)) {
 		    /* read from the alternate channel, signal the child */
-		    if ((*des_read)(s, &sig, 1) <= 0) {
+		    if (rcmd_stream_read(s, &sig, 1, 1) <= 0) {
 			FD_CLR(s, &readfrom);
 		    } else {
 #ifdef POSIX_SIGNALS
@@ -1233,7 +1277,7 @@ if(port)
 		if (FD_ISSET(f, &ready)) {
 		    /* read from the net, write to child stdin */
 		    errno = 0;
-		    cc = (*des_read)(f, buf, sizeof(buf));
+		    cc = rcmd_stream_read(f, buf, sizeof(buf), 0);
 		    if (cc <= 0) {
 			(void) close(px[1]);
 			FD_CLR(f, &readfrom);
@@ -1313,6 +1357,13 @@ if(port)
         initgroups(pwd->pw_name, pwd->pw_gid);
     }
 #endif
+#ifdef	HAVE_SETLUID
+    /*
+     * If we're on a system which keeps track of login uids, then
+     * set the login uid. 
+     */
+    setluid((uid_t) pwd->pw_uid);
+#endif	/* HAVE_SETLUID */
     (void) setuid((uid_t)pwd->pw_uid);
     /* if TZ is set in the parent, drag it in */
     {
@@ -1354,14 +1405,22 @@ if(port)
 
     {
       int i;
-      /* these two are covered by ADDRPAD */
+      /* these four are covered by ADDRPAD */
       sprintf(local_addr,  "KRB5LOCALADDR=%s", inet_ntoa(localaddr.sin_addr));
       for (i = 0; envinit[i]; i++);
       envinit[i] =local_addr;
 
+      sprintf(local_port,  "KRB5LOCALPORT=%d", ntohs(localaddr.sin_port));
+      for (; envinit[i]; i++);
+      envinit[i] =local_port;
+
       sprintf(remote_addr, "KRB5REMOTEADDR=%s", inet_ntoa(fromp->sin_addr));
       for (; envinit[i]; i++);
       envinit[i] =remote_addr;
+
+      sprintf(remote_port, "KRB5REMOTEPORT=%d", ntohs(fromp->sin_port));
+      for (; envinit[i]; i++);
+      envinit[i] =remote_port;
     }
 
     /* If we do anything else, make sure there is space in the array. */
@@ -1409,15 +1468,16 @@ if(port)
         strcpy((char *) cmdbuf + offst, kprogdir);
 	cp = copy + 3 + offst;
 
+	cmdbuf[sizeof(cmdbuf) - 1] = '\0';
 	if (auth_sys == KRB5_RECVAUTH_V4) {
-	  strcat(cmdbuf, "/v4rcp");
+	  strncat(cmdbuf, "/v4rcp", sizeof(cmdbuf) - 1 - strlen(cmdbuf));
 	} else {
-	  strcat(cmdbuf, "/rcp");
+	  strncat(cmdbuf, "/rcp", sizeof(cmdbuf) - 1 - strlen(cmdbuf));
 	}
 	if (stat((char *)cmdbuf + offst, &s) >= 0)
-	  strcat(cmdbuf, cp);
+	  strncat(cmdbuf, cp, sizeof(cmdbuf) - 1 - strlen(cmdbuf));
 	else
-	  strcpy(cmdbuf, copy);
+	  strncpy(cmdbuf, copy, sizeof(cmdbuf) - 1 - strlen(cmdbuf));
 	free(copy);
     }
 #endif
@@ -1453,7 +1513,7 @@ void error(fmt, a1, a2, a3)
      char *fmt;
      char *a1, *a2, *a3;
 {
-    char buf[RSHD_BUFSIZ];
+    char buf[RCMD_BUFSIZ];
     
     buf[0] = 1;
     (void) sprintf(buf+1, "%s: ", progname);
@@ -1714,14 +1774,17 @@ recvauth(netf, peersin, valid_checksum)
     char krb_vers[KRB_SENDAUTH_VLEN + 1];
     int len;
     krb5_data inbuf;
+#ifdef KRB5_KRB4_COMPAT
     char v4_instance[INST_SZ];	/* V4 Instance */
-    char v4_version[9];
+#endif
     krb5_authenticator *authenticator;
     krb5_ticket        *ticket;
     krb5_rcache		rcache;
     struct passwd *pwd;
     uid_t uid;
     gid_t gid;
+    enum kcmd_proto kcmd_proto;
+    krb5_data version;
 
     *valid_checksum = 0;
     len = sizeof(laddr);
@@ -1735,7 +1798,9 @@ recvauth(netf, peersin, valid_checksum)
 #define SIZEOF_INADDR sizeof(struct in_addr)
 #endif
 
+#ifdef KRB5_KRB4_COMPAT
     strcpy(v4_instance, "*");
+#endif
 
     if (status = krb5_auth_con_init(bsd_context, &auth_context))
 	return status;
@@ -1764,8 +1829,8 @@ recvauth(netf, peersin, valid_checksum)
 	if (status) return status;
     }
 
-    status = krb5_compat_recvauth(bsd_context, &auth_context, &netf,
-				  "KCMDV0.1",
+#ifdef KRB5_KRB4_COMPAT
+    status = krb5_compat_recvauth_version(bsd_context, &auth_context, &netf,
 				  NULL,		/* Specify daemon principal */
 				  0, 		/* no flags */
 				  keytab, /* normally NULL to use v5srvtab */
@@ -1778,7 +1843,16 @@ recvauth(netf, peersin, valid_checksum)
 
 				  &ticket, 	/* return ticket */
 				  &auth_sys, 	/* which authentication system*/
-				  &v4_kdata, 0, v4_version);
+				  &v4_kdata, 0, &version);
+#else
+    status = krb5_recvauth_version(bsd_context, &auth_context, &netf,
+				   NULL,        /* daemon principal */
+				   0,           /* no flags */
+				   keytab,      /* normally NULL to use v5srvtab */
+				   &ticket,    /* return ticket */
+				   &version); /* application version string */
+    auth_sys = KRB5_RECVAUTH_V5;
+#endif
 
     if (status) {
 	if (auth_sys == KRB5_RECVAUTH_V5) {
@@ -1795,7 +1869,10 @@ recvauth(netf, peersin, valid_checksum)
     getstr(netf, locuser, sizeof(locuser), "locuser");
     getstr(netf, cmdbuf, sizeof(cmdbuf), "command");
 
+#ifdef KRB5_KRB4_COMPAT
     if (auth_sys == KRB5_RECVAUTH_V4) {
+	rcmd_stream_init_normal();
+	
 	/* We do not really know the remote user's login name.
          * Assume it to be the same as the first component of the
 	 * principal's name. 
@@ -1811,9 +1888,18 @@ recvauth(netf, peersin, valid_checksum)
 	
 	return status;
     }
+#endif /* KRB5_KRB4_COMPAT */
 
     /* Must be V5  */
 	
+    kcmd_proto = KCMD_UNKNOWN_PROTOCOL;
+    if (version.length != 9)
+	fatal (netf, "bad application version length");
+    if (!memcmp (version.data, "KCMDV0.1", 9))
+	kcmd_proto = KCMD_OLD_PROTOCOL;
+    if (!memcmp (version.data, "KCMDV0.2", 9))
+	kcmd_proto = KCMD_NEW_PROTOCOL;
+
     getstr(netf, remuser, sizeof(locuser), "remuser");
 
     if ((status = krb5_unparse_name(bsd_context, ticket->enc_part2->client, 
@@ -1850,7 +1936,7 @@ recvauth(netf, peersin, valid_checksum)
 
     error_cleanup:
 	if (chksumbuf)
-	    krb5_xfree(chksumbuf);
+	    free(chksumbuf);
 	if (status) {
 	    krb5_free_authenticator(bsd_context, authenticator);
 	    return status;
@@ -1859,14 +1945,26 @@ recvauth(netf, peersin, valid_checksum)
     }
     krb5_free_authenticator(bsd_context, authenticator);
 
-    
-    /* Setup eblock for encrypted sessions. */
-    krb5_use_enctype(bsd_context, &eblock, ticket->enc_part2->session->enctype);
-    if ((status = krb5_process_key(bsd_context, &eblock,
-				   ticket->enc_part2->session)))
-	fatal(netf, "Permission denied");
 
-    /* Null out the "session" because eblock.key references the session
+    if (!strncmp(cmdbuf, "-x ", 3))
+	do_encrypt = 1;
+
+    {
+	krb5_keyblock *key;
+	status = krb5_auth_con_getremotesubkey (bsd_context, auth_context,
+						&key);
+	if (status)
+	    fatal (netf, "Server can't get session subkey");
+	if (!key && do_encrypt && kcmd_proto == KCMD_NEW_PROTOCOL)
+	    fatal (netf, "No session subkey sent");
+	if (key && kcmd_proto == KCMD_OLD_PROTOCOL)
+	    fatal (netf, "Session subkey not allowed in old kcmd protocol");
+	if (key == 0)
+	    key = ticket->enc_part2->session;
+	rcmd_stream_init_krb5 (key, do_encrypt, 0, 0, kcmd_proto);
+    }
+
+    /* Null out the "session" because kcmd.c references the session
      * key here, and we do not want krb5_free_ticket() to destroy it. */
     ticket->enc_part2->session = 0;
 
@@ -1899,137 +1997,9 @@ recvauth(netf, peersin, valid_checksum)
     krb5_free_ticket(bsd_context, ticket);
     return 0;
 }
+#endif /* KERBEROS */
 
 
-char storage[2*RSHD_BUFSIZ];               /* storage for the decryption */
-int nstored = 0;
-char *store_ptr = storage;
-
-int
-v5_des_read(fd, buf, len)
-     int fd;
-     register char *buf;
-     int len;
-{
-    int nreturned = 0;
-    krb5_ui_4 net_len,rd_len;
-    int cc,retry;
-    unsigned char len_buf[4];
-    
-    if (!do_encrypt)
-      return(read(fd, buf, len));
-    
-    if (nstored >= len) {
-	memcpy(buf, store_ptr, len);
-	store_ptr += len;
-	nstored -= len;
-	return(len);
-    } else if (nstored) {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	buf += nstored;
-	len -= nstored;
-	nstored = 0;
-    }
-    
-    if ((cc = krb5_net_read(bsd_context, fd, (char *)len_buf, 4)) != 4) {
-	if ((cc < 0)  && ((errno == EWOULDBLOCK) || (errno == EAGAIN)))
-	    return(cc);
-	/* XXX can't read enough, pipe must have closed */
-	return(0);
-    }
-    rd_len =
-	((len_buf[0]<<24) |
-	 (len_buf[1]<<16) |
-	 (len_buf[2]<<8) |
-	 len_buf[3]);
-    net_len = krb5_encrypt_size(rd_len, eblock.crypto_entry);
-    /* note: net_len is unsigned */
-    if (net_len > sizeof(des_inbuf)) {
-	/* XXX preposterous length, probably out of sync.
-	   act as if pipe closed */
-	syslog(LOG_ERR,"Read size problem (rd_len=%d, net_len=%d)",
-	       rd_len, net_len);
-	return(0);
-    }
-    retry = 0;
-  datard:
-    if ((cc = krb5_net_read(bsd_context, fd, desinbuf.data, net_len)) != net_len) {
-	/* XXX can't read enough, pipe must have closed */
-	if ((cc < 0)  && ((errno == EWOULDBLOCK) || (errno == EAGAIN))) {
-	    retry++;
-	    if (retry > MAXRETRIES){
-		syslog(LOG_ERR, "des_read retry count exceeded %d\n", retry);
-		return(0);
-	    }
-	    sleep(1);
-	    goto datard;
-	}
-	syslog(LOG_ERR,
-	       "Read data received %d != expected %d.",
-	       cc, net_len);
-	return(0);
-    }
-
-    /* decrypt info */
-    if (krb5_decrypt(bsd_context, desinbuf.data, (krb5_pointer) storage, net_len,
-		     &eblock, 0)) {
-	syslog(LOG_ERR,"Read decrypt problem.");
-	return(0);
-    }
-
-    store_ptr = storage;
-    nstored = rd_len;
-    if (nstored > len) {
-	memcpy(buf, store_ptr, len);
-	nreturned += len;
-	store_ptr += len;
-	nstored -= len;
-    } else {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	nstored = 0;
-    }
-    return(nreturned);
-}
-    
-
-int
-v5_des_write(fd, buf, len)
-     int fd;
-     char *buf;
-     int len;
-{
-    unsigned char len_buf[4];
-    
-    if (!do_encrypt)
-      return(write(fd, buf, len));
-    
-    desoutbuf.length = krb5_encrypt_size(len, eblock.crypto_entry);
-    if (desoutbuf.length > sizeof(des_outbuf)){
-	syslog(LOG_ERR,"Write size problem (%d > %d)",
-	       desoutbuf.length, sizeof(des_outbuf));
-	return(-1);
-    }
-
-    if (krb5_encrypt(bsd_context, (krb5_pointer)buf, desoutbuf.data, len, &eblock, 0)) {
-	syslog(LOG_ERR,"Write encrypt problem.");
-	return(-1);
-    }
-    
-    len_buf[0] = (len & 0xff000000) >> 24;
-    len_buf[1] = (len & 0xff0000) >> 16;
-    len_buf[2] = (len & 0xff00) >> 8;
-    len_buf[3] = (len & 0xff);
-    (void) write(fd, len_buf, 4);
-
-    if (write(fd, desoutbuf.data, desoutbuf.length) != desoutbuf.length){
-	syslog(LOG_ERR,"Could not write out all data.");
-	return(-1);
-    }
-    else return(len);
-}
-
 
 void fatal(f, msg)
      int f;
@@ -2043,7 +2013,7 @@ void fatal(f, msg)
     buf[0] = '\01';             /* error indicator */
     (void) sprintf(buf + 1, "%s: %s.\r\n",progname, msg);
     if ((f == netf) && (pid > 0))
-      (void) (*des_write)(f, buf, strlen(buf));
+      (void) rcmd_stream_write(f, buf, strlen(buf), 0);
     else
       (void) write(f, buf, strlen(buf));
     syslog(LOG_ERR,"%s\n",msg);
@@ -2059,4 +2029,3 @@ void fatal(f, msg)
     }
     exit(1);
 }
-#endif /* KERBEROS */
