@@ -14,7 +14,10 @@
  * this permission notice appear in supporting documentation, and that
  * the name of M.I.T. not be used in advertising or publicity pertaining
  * to distribution of the software without specific, written prior
- * permission.  M.I.T. makes no representations about the suitability of
+ * permission.  Furthermore if you modify this software you must label
+ * your software as modified software and not distribute it in such a
+ * fashion that it might be confused with the original M.I.T. software.
+ * M.I.T. makes no representations about the suitability of
  * this software for any purpose.  It is provided "as is" without express
  * or implied warranty.
  * 
@@ -55,7 +58,8 @@ static struct pflag flags[] = {
 {"requires_hwauth",	15,	KRB5_KDB_REQUIRES_HW_AUTH,	0},
 {"needchange",		10,	KRB5_KDB_REQUIRES_PWCHANGE,	0},
 {"allow_svr",		9,	KRB5_KDB_DISALLOW_SVR,		1},
-{"password_changing_service",	25,	KRB5_KDB_PWCHANGE_SERVICE,	0 }
+{"password_changing_service",	25,	KRB5_KDB_PWCHANGE_SERVICE,	0 },
+{"support_desmd5",	14,	KRB5_KDB_SUPPORT_DESMD5,	0 }
 };
 
 static char *prflags[] = {
@@ -72,7 +76,9 @@ static char *prflags[] = {
     "UNKNOWN_0x00000400",	/* 0x00000400 */
     "UNKNOWN_0x00000800",	/* 0x00000800 */
     "DISALLOW_SVR",		/* 0x00001000 */
-    "PWCHANGE_SERVICE"		/* 0x00002000 */
+    "PWCHANGE_SERVICE",		/* 0x00002000 */
+    "SUPPORT_DESMD5",		/* 0x00004000 */
+    "NEW_PRINC",		/* 0x00008000 */
 };
 
 char *getenv();
@@ -176,7 +182,6 @@ char *kadmin_startup(argc, argv)
 	 com_err(whoami, retval, "while initializing krb5 library");
 	 exit(1);
     }
-    krb5_init_ets(context);
 		     
     while ((optchar = getopt(argc, argv, "r:p:kq:w:d:s:mc:t:e:")) != EOF) {
 	switch (optchar) {
@@ -284,7 +289,7 @@ char *kadmin_startup(argc, argv)
 	     }
 	     krb5_free_principal(context, princ);
 	     freeprinc++;
-	} else if (use_keytab != NULL) {
+	} else if (use_keytab != 0) {
 	     if (retval = krb5_sname_to_principal(context, NULL,
 						  "host",
 						  KRB5_NT_SRV_HST,
@@ -377,27 +382,38 @@ char *kadmin_startup(argc, argv)
      * Initialize the kadm5 connection.  If we were given a ccache,
      * use it.  Otherwise, use/prompt for the password.
      */
-    if (ccache_name)
+    if (ccache_name) {
+	 printf("Authenticating as principal %s with existing credentials.\n",
+		princstr);
 	 retval = kadm5_init_with_creds(princstr, cc,
 					KADM5_ADMIN_SERVICE, 
 					&params,
 					KADM5_STRUCT_VERSION,
 					KADM5_API_VERSION_2,
 					&handle);
-    else if (use_keytab)
+    } else if (use_keytab) {
+	 if (keytab_name)
+	     printf("Authenticating as principal %s with keytab %s.\n",
+		    princstr, keytab_name);
+	 else
+	     printf("Authenticating as principal %s with default keytab.\n",
+		    princstr);
 	 retval = kadm5_init_with_skey(princstr, keytab_name,
 				       KADM5_ADMIN_SERVICE, 
 				       &params,
 				       KADM5_STRUCT_VERSION,
 				       KADM5_API_VERSION_2,
 				       &handle);
-    else
+    } else {
+	 printf("Authenticating as principal %s with password.\n",
+		princstr);
 	 retval = kadm5_init_with_password(princstr, password,
 					   KADM5_ADMIN_SERVICE, 
 					   &params,
 					   KADM5_STRUCT_VERSION,
 					   KADM5_API_VERSION_2,
 					   &handle);
+    }
     if (retval) {
 	com_err(whoami, retval, "while initializing %s interface", whoami);
 	if (retval == KADM5_BAD_CLIENT_PARAMS ||
@@ -443,6 +459,7 @@ int quit()
      }
 
      /* insert more random cleanup here */
+     krb5_free_context(context);
      return 0;
 }
 
@@ -503,25 +520,74 @@ void kadmin_cpw(argc, argv)
     static char newpw[1024];
     static char prompt1[1024], prompt2[1024];
     char *canon;
+    char *pwarg = NULL;
+    int n_ks_tuple = 0, keepold = 0, randkey = 0;
+    krb5_key_salt_tuple *ks_tuple = NULL;
     krb5_principal princ;
     
     if (argc < 2) {
 	 goto usage;
     }
-    
-    retval = kadmin_parse_name(argv[argc - 1], &princ);
+    for (argv++, argc--; argc > 1; argc--, argv++) {
+	if (!strcmp("-pw", *argv)) {
+	    argc--;
+	    if (argc < 1) {
+		fprintf(stderr, "change_password: missing password arg\n");
+		goto usage;
+	    }
+	    pwarg = *++argv;
+	    continue;
+	}
+	if (!strcmp("-randkey", *argv)) {
+	    randkey++;
+	    continue;
+	}
+	if (!strcmp("-keepold", *argv)) {
+	    keepold++;
+	    continue;
+	}
+	if (!strcmp("-e", *argv)) {
+	    argc--;
+	    if (argc < 1) {
+		fprintf(stderr,
+			"change_password: missing keysaltlist arg\n");
+		goto usage;
+	    }
+	    retval = krb5_string_to_keysalts(*++argv, ", \t", ":.-", 0,
+					     &ks_tuple, &n_ks_tuple);
+	    if (retval) {
+		com_err("change_password", retval,
+			"while parsing keysalts %s", *argv);
+		return;
+	    }
+	    continue;
+	}
+	goto usage;
+    }
+    retval = kadmin_parse_name(*argv, &princ);
     if (retval) {
 	com_err("change_password", retval, "while parsing principal name");
+	if (ks_tuple != NULL)
+	    free(ks_tuple);
 	return;
     }
     retval = krb5_unparse_name(context, princ, &canon);
     if (retval) {
 	com_err("change_password", retval, "while canonicalizing principal");
 	krb5_free_principal(context, princ);
+	if (ks_tuple != NULL)
+	    free(ks_tuple);
 	return;
     }
-    if ((argc == 4) && (strlen(argv[1]) == 3) && !strcmp("-pw", argv[1])) {
-	retval = kadm5_chpass_principal(handle, princ, argv[2]);
+    if (pwarg != NULL) {
+	if (keepold || ks_tuple != NULL) {
+	    retval = kadm5_chpass_principal_3(handle, princ, keepold,
+					      n_ks_tuple, ks_tuple, pwarg);
+	    if (ks_tuple != NULL)
+		free(ks_tuple);
+	} else {
+	    retval = kadm5_chpass_principal(handle, princ, pwarg);
+	}
 	krb5_free_principal(context, princ);
 	if (retval) {
 	    com_err("change_password", retval,
@@ -532,9 +598,16 @@ void kadmin_cpw(argc, argv)
 	printf("Password for \"%s\" changed.\n", canon);
 	free(canon);
 	return;
-    } else if ((argc == 3) && (strlen(argv[1]) == 8) &&
-	       !strcmp("-randkey", argv[1])) {
-	retval = kadm5_randkey_principal(handle, princ, NULL, NULL);
+    } else if (randkey) {
+	if (keepold || ks_tuple != NULL) {
+	    retval = kadm5_randkey_principal_3(handle, princ, keepold,
+					       n_ks_tuple, ks_tuple,
+					       NULL, NULL);
+	    if (ks_tuple != NULL)
+		free(ks_tuple);
+	} else {
+	    retval = kadm5_randkey_principal(handle, princ, NULL, NULL);
+	}
 	krb5_free_principal(context, princ);
 	if (retval) {
 	    com_err("change_password", retval,
@@ -545,24 +618,34 @@ void kadmin_cpw(argc, argv)
 	printf("Key for \"%s\" randomized.\n", canon);
 	free(canon);
 	return;
-    } else if (argc == 2) {
+    } else if (argc == 1) {
 	int i = sizeof (newpw) - 1;
 	
 	sprintf(prompt1, "Enter password for principal \"%.900s\": ",
-		argv[1]);
+		*argv);
 	sprintf(prompt2,
 		"Re-enter password for principal \"%.900s\": ",
-		argv[1]);
+		*argv);
 	retval = krb5_read_password(context, prompt1, prompt2,
 				    newpw, &i);
 	if (retval) {
 	    com_err("change_password", retval,
 		    "while reading password for \"%s\".", canon);
 	    free(canon);
+	    if (ks_tuple != NULL)
+		free(ks_tuple);
 	    krb5_free_principal(context, princ);
 	    return;
 	}
-	retval = kadm5_chpass_principal(handle, princ, newpw);
+	if (keepold || ks_tuple != NULL) {
+	    retval = kadm5_chpass_principal_3(handle, princ, keepold,
+					      n_ks_tuple, ks_tuple,
+					      newpw);
+	    if (ks_tuple != NULL)
+		free(ks_tuple);
+	} else {
+	    retval = kadm5_chpass_principal(handle, princ, newpw);
+	}
 	krb5_free_principal(context, princ);
 	memset(newpw, 0, sizeof (newpw));
 	if (retval) {
@@ -578,20 +661,26 @@ void kadmin_cpw(argc, argv)
 	free(canon);
 	krb5_free_principal(context, princ);
    usage:
+	if (ks_tuple != NULL)
+	    free(ks_tuple);
 	fprintf(stderr,
-		"usage: change_password [-randkey] [-pw password] "
+		"usage: change_password [-randkey] [-keepold] "
+		"[-e keysaltlist] [-pw password] "
 		"principal\n");
 	return;
    }
 }
 
-int kadmin_parse_princ_args(argc, argv, oprinc, mask, pass, randkey, caller)
+int kadmin_parse_princ_args(argc, argv, oprinc, mask, pass, randkey,
+			    ks_tuple, n_ks_tuple, caller)
     int argc;
     char *argv[];
     kadm5_principal_ent_t oprinc;
     long *mask;
     char **pass;
     int *randkey;
+    krb5_key_salt_tuple **ks_tuple;
+    int *n_ks_tuple;
     char *caller;
 {
     int i, j, attrib_set;
@@ -601,6 +690,8 @@ int kadmin_parse_princ_args(argc, argv, oprinc, mask, pass, randkey, caller)
     
     *mask = 0;
     *pass = NULL;
+    *n_ks_tuple = 0;
+    *ks_tuple = NULL;
     time(&now);
     *randkey = 0;
     for (i = 1; i < argc - 1; i++) {
@@ -709,6 +800,20 @@ int kadmin_parse_princ_args(argc, argv, oprinc, mask, pass, randkey, caller)
 	    ++*randkey;
 	    continue;
 	}
+	if (!strcmp("-e", argv[i])) {
+	    if (++i > argc - 2)
+		return -1;
+	    else {
+		retval = krb5_string_to_keysalts(argv[i], ", \t", ":.-", 0,
+						 ks_tuple, n_ks_tuple);
+		if (retval) {
+		    com_err(caller, retval,
+			    "while parsing keysalts %s", argv[i]);
+		    return -1;
+		}
+	    }
+	    continue;
+	}
 	for (j = 0; j < sizeof (flags) / sizeof (struct pflag); j++) {
 	    if (strlen(argv[i]) == flags[j].flaglen + 1 &&
 		!strcmp(flags[j].flagname,
@@ -749,7 +854,7 @@ void kadmin_addprinc_usage(func)
 {
      fprintf(stderr, "usage: %s [options] principal\n", func);
      fprintf(stderr, "\toptions are:\n");
-     fprintf(stderr, "\t\t[-expire expdate] [-pwexpire pwexpdate] [-maxlife maxtixlife]\n\t\t[-kvno kvno] [-policy policy] [-randkey] [-pw password]\n\t\t[-maxrenewlife maxrenewlife] [{+|-}attribute]\n");
+     fprintf(stderr, "\t\t[-expire expdate] [-pwexpire pwexpdate] [-maxlife maxtixlife]\n\t\t[-kvno kvno] [-policy policy] [-randkey] [-pw password]\n\t\t[-maxrenewlife maxrenewlife]\n\t\t[-e keysaltlist]\n\t\t[{+|-}attribute]\n");
      fprintf(stderr, "\tattributes are:\n");
      fprintf(stderr, "%s%s%s",
 	     "\t\tallow_postdated allow_forwardable allow_tgs_req allow_renewable\n",
@@ -778,6 +883,8 @@ void kadmin_addprinc(argc, argv)
     kadm5_policy_ent_rec defpol;
     long mask;
     int randkey = 0, i;
+    int n_ks_tuple;
+    krb5_key_salt_tuple *ks_tuple;
     char *pass, *canon;
     krb5_error_code retval;
     static char newpw[1024], dummybuf[256];
@@ -794,6 +901,7 @@ void kadmin_addprinc(argc, argv)
     princ.attributes = 0;
     if (kadmin_parse_princ_args(argc, argv,
 				&princ, &mask, &pass, &randkey,
+				&ks_tuple, &n_ks_tuple,
 				"add_principal")) {
 	 kadmin_addprinc_usage("add_principal");
 	 return;
@@ -804,6 +912,8 @@ void kadmin_addprinc(argc, argv)
 	com_err("add_principal",
 		retval, "while canonicalizing principal");
 	krb5_free_principal(context, princ.principal);
+	if (ks_tuple != NULL)
+	    free(ks_tuple);
 	return;
     }
 
@@ -814,11 +924,18 @@ void kadmin_addprinc(argc, argv)
      * unset, since it is never valid for kadm5_create_principal.
      */
     if ((! (mask & KADM5_POLICY)) &&
-	(! (mask & KADM5_POLICY_CLR)) &&
-	(! (retval = kadm5_get_policy(handle, "default", &defpol)))) {
-	 princ.policy = "default";
-	 mask |= KADM5_POLICY;
-	 (void) kadm5_free_policy_ent(handle, &defpol);
+	(! (mask & KADM5_POLICY_CLR))) {
+	 if (! kadm5_get_policy(handle, "default", &defpol)) {
+	      fprintf(stderr,
+		"NOTICE: no policy specified for %s; assigning \"default\"\n",
+		      canon);
+	      princ.policy = "default";
+	      mask |= KADM5_POLICY;
+	      (void) kadm5_free_policy_ent(handle, &defpol);
+	 } else
+	      fprintf(stderr,
+	     "WARNING: no policy specified for %s; defaulting to no policy\n",
+		      canon);
     }
     mask &= ~KADM5_POLICY_CLR;
     
@@ -846,22 +963,38 @@ void kadmin_addprinc(argc, argv)
 	pass = newpw;
     }
     mask |= KADM5_PRINCIPAL;
-    retval = kadm5_create_principal(handle, &princ, mask, pass);
+    if (ks_tuple != NULL) {
+	retval = kadm5_create_principal_3(handle, &princ, mask,
+					  n_ks_tuple, ks_tuple, pass);
+    } else {
+	retval = kadm5_create_principal(handle, &princ, mask, pass);
+    }
     if (retval) {
 	com_err("add_principal", retval, "while creating \"%s\".",
 		canon);
 	krb5_free_principal(context, princ.principal);
 	free(canon);
+	if (ks_tuple != NULL)
+	    free(ks_tuple);
 	return;
     }
     if (randkey) {		/* more special stuff for -randkey */
-	retval = kadm5_randkey_principal(handle, princ.principal,
-					 NULL, NULL);
+	if (ks_tuple != NULL) {
+	    retval = kadm5_randkey_principal_3(handle, princ.principal,
+					       FALSE,
+					       n_ks_tuple, ks_tuple,
+					       NULL, NULL);
+	} else {
+	    retval = kadm5_randkey_principal(handle, princ.principal,
+					     NULL, NULL);
+	}
 	if (retval) {
 	    com_err("add_principal", retval,
 		    "while randomizing key for \"%s\".", canon);
 	    krb5_free_principal(context, princ.principal);
 	    free(canon);
+	    if (ks_tuple != NULL)
+		free(ks_tuple);
 	    return;
 	}
 	princ.attributes &= ~KRB5_KDB_DISALLOW_ALL_TIX;	/* clear notix */
@@ -872,11 +1005,15 @@ void kadmin_addprinc(argc, argv)
 		    "while clearing DISALLOW_ALL_TIX for \"%s\".", canon);
 	    krb5_free_principal(context, princ.principal);
 	    free(canon);
+	    if (ks_tuple != NULL)
+		free(ks_tuple);
 	    return;
 	}
     }
     krb5_free_principal(context, princ.principal);
     printf("Principal \"%s\" created.\n", canon);
+    if (ks_tuple != NULL)
+	free(ks_tuple);
     free(canon);
 }
 
@@ -890,6 +1027,8 @@ void kadmin_modprinc(argc, argv)
     krb5_error_code retval;
     char *pass, *canon;
     int randkey = 0;
+    int n_ks_tuple = 0;
+    krb5_key_salt_tuple *ks_tuple;
 
     if (argc < 2) {
 	 kadmin_modprinc_usage("modify_principal");
@@ -925,7 +1064,14 @@ void kadmin_modprinc(argc, argv)
     retval = kadmin_parse_princ_args(argc, argv,
 				     &princ, &mask,
 				     &pass, &randkey,
+				     &ks_tuple, &n_ks_tuple,
 				     "modify_principal");
+    if (ks_tuple != NULL) {
+	free(ks_tuple);
+	kadmin_modprinc_usage("modify_principal");
+	free(canon);
+	return;
+    }
     if (retval) {
 	kadmin_modprinc_usage("modify_principal");
 	free(canon);
@@ -937,6 +1083,13 @@ void kadmin_modprinc(argc, argv)
 	free(canon);
 	return;
     }
+    if (pass) {
+	fprintf(stderr,
+		"modify_principal: -pw not allowed; use change_password\n");
+	krb5_free_principal(context, princ.principal);
+	free(canon);
+	return;
+    }
     retval = kadm5_modify_principal(handle, &princ, mask);
     krb5_free_principal(context, princ.principal);
     if (retval) {
