@@ -12,4 +12,4 @@
 
 /* Format: "KRB5_BRAND: <cvs tag> <date>" */
 
-static char krb5_brand[] = "KRB5_BRAND: krb5-1-2-final 1.2 20000623";
+static char krb5_brand[] = "KRB5_BRAND: krb5-1-2-1-final 1.2.1 20000629";
